(function() {
  var chai, invalidExamples, schemas, testInvalid, testValid, tv4, validExamples;

  chai = require('chai');

  schemas = require('../../schema/schemas.js');

  tv4 = require('tv4');

  validExamples = {
    'start recording': {
      protocol: 'trace',
      command: 'start',
      payload: {
        graph: 'mygraph',
        secret: 'verygood'
      }
    },
    'dump flowtrace.json': {
      protocol: 'trace',
      command: 'dump',
      payload: {
        graph: 'mygraph',
        secret: 'verygood',
        type: 'flowtrace.json'
      }
    },
    'clear buffer': {
      protocol: 'trace',
      command: 'clear',
      payload: {
        graph: 'mygraph',
        secret: 'verygood'
      }
    },
    'stop recording': {
      protocol: 'trace',
      command: 'stop',
      payload: {
        graph: 'mygraph',
        secret: 'verygood'
      }
    }
  };

  invalidExamples = {
    'start with invalid graph id': {
      protocol: 'trace',
      command: 'start',
      payload: {
        graph: 112.0,
        secret: 'nnice'
      }
    },
    'dump with invalid type': {
      protocol: 'trace',
      command: 'dump',
      payload: {
        graph: 'mygraph',
        secret: 'verygood',
        type: 'not-a-valid-trace-type'
      }
    },
    'stop with missing secret': {
      protocol: 'trace',
      command: 'start',
      payload: {
        graph: 112.0,
        secret: void 0
      }
    },
    'clear without graph': {
      protocol: 'trace',
      command: 'start',
      payload: {
        graph: void 0,
        secret: 'verygood'
      }
    }
  };

  testValid = function(name) {
    return describe(name, function() {
      return it('should validate OK', function() {
        var data, result, schema, schemaId;
        data = validExamples[name];
        schemaId = "/" + data.protocol + "/input/" + data.command;
        schema = tv4.getSchema(schemaId);
        result = tv4.validateMultiple(data, schema);
        chai.expect(result.missing, "missing schemas " + result.missing).to.eql([]);
        return chai.expect(result.errors, result.errors).to.eql([]);
      });
    });
  };

  testInvalid = function(name) {
    return describe(name, function() {
      return it('should error', function() {
        var data, result, schema, schemaId;
        data = invalidExamples[name];
        schemaId = "/" + data.protocol + "/input/" + data.command;
        schema = tv4.getSchema(schemaId);
        result = tv4.validateMultiple(data, schema);
        return chai.expect(result.valid).to.be["false"];
      });
    });
  };

  describe('Tracing protocol schema', function() {
    var traceSchema;
    traceSchema = null;
    before(function() {
      var sharedSchema;
      sharedSchema = schemas.shared;
      traceSchema = schemas.trace;
      tv4.addSchema('/shared/', sharedSchema);
      return tv4.addSchema('/trace/', traceSchema);
    });
    it('should exist', function() {
      chai.expect(traceSchema).to.exist;
      chai.expect(traceSchema).to.be.an('object');
      chai.expect(traceSchema.input).to.be.an('object');
      return chai.expect(traceSchema.output).to.be.an('object');
    });
    describe('valid examples', function() {
      return Object.keys(validExamples).forEach(function(k) {
        return testValid(k);
      });
    });
    return describe('invalid examples', function() {
      return Object.keys(invalidExamples).forEach(function(k) {
        return testInvalid(k);
      });
    });
  });

}).call(this);
